package java.lang.invoke;

import sun.invoke.util.BytecodeDescriptor;
import jdk.internal.org.objectweb.asm.Type;
import sun.invoke.util.Wrapper;
import jdk.internal.org.objectweb.asm.MethodVisitor;

class TypeConvertingMethodAdapter extends MethodVisitor
{
    private static final int NUM_WRAPPERS;
    private static final String NAME_OBJECT = "java/lang/Object";
    private static final String WRAPPER_PREFIX = "Ljava/lang/";
    private static final String NAME_BOX_METHOD = "valueOf";
    private static final int[][] wideningOpcodes;
    private static final Wrapper[] FROM_WRAPPER_NAME;
    private static final Wrapper[] FROM_TYPE_SORT;
    
    TypeConvertingMethodAdapter(MethodVisitor methodVisitor) {
        super(327680, methodVisitor);
    }
    
    private static void initWidening(Wrapper wrapper, int n, Wrapper... array) {
        for (int length = array.length, i = 0; i < length; ++i) {
            TypeConvertingMethodAdapter.wideningOpcodes[array[i].ordinal()][wrapper.ordinal()] = n;
        }
    }
    
    private static int hashWrapperName(String s) {
        if (s.length() < 3) {
            return 0;
        }
        return (3 * s.charAt(1) + s.charAt(2)) % 16;
    }
    
    private Wrapper wrapperOrNullFromDescriptor(String s) {
        if (!s.startsWith("Ljava/lang/")) {
            return null;
        }
        String substring = s.substring("Ljava/lang/".length(), s.length() - 1);
        Wrapper wrapper = TypeConvertingMethodAdapter.FROM_WRAPPER_NAME[hashWrapperName(substring)];
        if (wrapper == null || wrapper.wrapperSimpleName().equals(substring)) {
            return wrapper;
        }
        return null;
    }
    
    private static String wrapperName(Wrapper wrapper) {
        return "java/lang/" + wrapper.wrapperSimpleName();
    }
    
    private static String unboxMethod(Wrapper wrapper) {
        return wrapper.primitiveSimpleName() + "Value";
    }
    
    private static String boxingDescriptor(Wrapper wrapper) {
        return String.format("(%s)L%s;", new Object[] { Character.valueOf(wrapper.basicTypeChar()), wrapperName(wrapper) });
    }
    
    private static String unboxingDescriptor(Wrapper wrapper) {
        return "()" + wrapper.basicTypeChar();
    }
    
    void boxIfTypePrimitive(Type type) {
        Wrapper wrapper = TypeConvertingMethodAdapter.FROM_TYPE_SORT[type.getSort()];
        if (wrapper != null) {
            this.box(wrapper);
        }
    }
    
    void widen(Wrapper wrapper, Wrapper wrapper2) {
        if (wrapper != wrapper2) {
            int opcode = TypeConvertingMethodAdapter.wideningOpcodes[wrapper.ordinal()][wrapper2.ordinal()];
            if (opcode != 0) {
                this.visitInsn(opcode);
            }
        }
    }
    
    void box(Wrapper wrapper) {
        this.visitMethodInsn(184, wrapperName(wrapper), "valueOf", boxingDescriptor(wrapper), false);
    }
    
    void unbox(String owner, Wrapper wrapper) {
        this.visitMethodInsn(182, owner, unboxMethod(wrapper), unboxingDescriptor(wrapper), false);
    }
    
    private String descriptorToName(String s) {
        int n = s.length() - 1;
        if (s.charAt(0) == 'L' && s.charAt(n) == ';') {
            return s.substring(1, n);
        }
        return s;
    }
    
    void cast(String s, String s2) {
        String descriptorToName = this.descriptorToName(s);
        String descriptorToName2 = this.descriptorToName(s2);
        if (!descriptorToName2.equals(descriptorToName) && !descriptorToName2.equals("java/lang/Object")) {
            this.visitTypeInsn(192, descriptorToName2);
        }
    }
    
    private boolean isPrimitive(Wrapper wrapper) {
        return wrapper != Wrapper.OBJECT;
    }
    
    private Wrapper toWrapper(String s) {
        int char1 = s.charAt(0);
        if (char1 == '[' || char1 == '(') {
            char1 = 'L';
        }
        return Wrapper.forBasicType((char)char1);
    }
    
    void convertType(Class<?> clazz, Class<?> clazz2, Class<?> clazz3) {
        if (clazz.equals(clazz2) && clazz.equals(clazz3)) {
            return;
        }
        if (clazz == Void.TYPE || clazz2 == Void.TYPE) {
            return;
        }
        if (clazz.isPrimitive()) {
            Wrapper forPrimitiveType = Wrapper.forPrimitiveType(clazz);
            if (clazz2.isPrimitive()) {
                this.widen(forPrimitiveType, Wrapper.forPrimitiveType(clazz2));
            }
            else {
                String unparse2 = BytecodeDescriptor.unparse(clazz2);
                Wrapper wrapperOrNullFromDescriptor2 = this.wrapperOrNullFromDescriptor(unparse2);
                if (wrapperOrNullFromDescriptor2 != null) {
                    this.widen(forPrimitiveType, wrapperOrNullFromDescriptor2);
                    this.box(wrapperOrNullFromDescriptor2);
                }
                else {
                    this.box(forPrimitiveType);
                    this.cast(wrapperName(forPrimitiveType), unparse2);
                }
            }
        }
        else {
            String s = BytecodeDescriptor.unparse(clazz);
            if (clazz3.isPrimitive()) {
            }
            else {
                s = BytecodeDescriptor.unparse(clazz3);
                this.cast(s, s);
            }
            String unparse = BytecodeDescriptor.unparse(clazz2);
            if (clazz2.isPrimitive()) {
                Wrapper wrapper = this.toWrapper(unparse);
                Wrapper wrapperOrNullFromDescriptor = this.wrapperOrNullFromDescriptor(s);
                if (wrapperOrNullFromDescriptor != null) {
                    if (wrapperOrNullFromDescriptor.isSigned() || wrapperOrNullFromDescriptor.isFloating()) {
                        this.unbox(wrapperName(wrapperOrNullFromDescriptor), wrapper);
                    }
                    else {
                        this.unbox(wrapperName(wrapperOrNullFromDescriptor), wrapperOrNullFromDescriptor);
                        this.widen(wrapperOrNullFromDescriptor, wrapper);
                    }
                }
                else {
                    String wrapperName;
                    if (wrapper.isSigned() || wrapper.isFloating()) {
                        wrapperName = "java/lang/Number";
                    }
                    else {
                        wrapperName = wrapperName(wrapper);
                    }
                    this.cast(s, wrapperName);
                    this.unbox(wrapperName, wrapper);
                }
            }
            else {
                this.cast(s, unparse);
            }
        }
        return;
    }
    
    void iconst(int operand) {
        if (operand >= -1 && operand <= 5) {
            this.mv.visitInsn(3 + operand);
        }
        else if (operand >= -128 && operand <= 127) {
            this.mv.visitIntInsn(16, operand);
        }
        else if (operand >= -32768 && operand <= 32767) {
            this.mv.visitIntInsn(17, operand);
        }
        else {
            this.mv.visitLdcInsn(Integer.valueOf(operand));
        }
    }
    
    static {
        NUM_WRAPPERS = Wrapper.values().length;
        wideningOpcodes = new int[TypeConvertingMethodAdapter.NUM_WRAPPERS][TypeConvertingMethodAdapter.NUM_WRAPPERS];
        FROM_WRAPPER_NAME = new Wrapper[16];
        FROM_TYPE_SORT = new Wrapper[16];
        for (Wrapper wrapper : Wrapper.values()) {
            if (wrapper.basicTypeChar() != 'L') {
                int hashWrapperName = hashWrapperName(wrapper.wrapperSimpleName());
                assert TypeConvertingMethodAdapter.FROM_WRAPPER_NAME[hashWrapperName] == null;
                TypeConvertingMethodAdapter.FROM_WRAPPER_NAME[hashWrapperName] = wrapper;
            }
        }
        for (int j = 0; j < TypeConvertingMethodAdapter.NUM_WRAPPERS; ++j) {
            for (int k = 0; k < TypeConvertingMethodAdapter.NUM_WRAPPERS; ++k) {
                TypeConvertingMethodAdapter.wideningOpcodes[j][k] = 0;
            }
        }
        initWidening(Wrapper.LONG, 133, new Wrapper[] { Wrapper.BYTE, Wrapper.SHORT, Wrapper.INT, Wrapper.CHAR });
        initWidening(Wrapper.LONG, 140, new Wrapper[] { Wrapper.FLOAT });
        initWidening(Wrapper.FLOAT, 134, new Wrapper[] { Wrapper.BYTE, Wrapper.SHORT, Wrapper.INT, Wrapper.CHAR });
        initWidening(Wrapper.FLOAT, 137, new Wrapper[] { Wrapper.LONG });
        initWidening(Wrapper.DOUBLE, 135, new Wrapper[] { Wrapper.BYTE, Wrapper.SHORT, Wrapper.INT, Wrapper.CHAR });
        initWidening(Wrapper.DOUBLE, 141, new Wrapper[] { Wrapper.FLOAT });
        initWidening(Wrapper.DOUBLE, 138, new Wrapper[] { Wrapper.LONG });
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[3] = Wrapper.BYTE;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[4] = Wrapper.SHORT;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[5] = Wrapper.INT;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[7] = Wrapper.LONG;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[2] = Wrapper.CHAR;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[6] = Wrapper.FLOAT;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[8] = Wrapper.DOUBLE;
        TypeConvertingMethodAdapter.FROM_TYPE_SORT[1] = Wrapper.BOOLEAN;
    }
}
